@@ -16,15 +16,16 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+from gi.repository import GObject
 from gi.repository import Gtk
 
 from pynicotine.config import config
 from pynicotine.core import core
 from pynicotine.gtkgui.application import GTK_API_VERSION
+from pynicotine.gtkgui.widgets.combobox import ComboBox
 from pynicotine.gtkgui.widgets.dialogs import Dialog
 from pynicotine.gtkgui.widgets.dialogs import EntryDialog
 from pynicotine.gtkgui.widgets.filechooser import FileChooserButton
-from pynicotine.gtkgui.widgets.textentry import ComboBox
 from pynicotine.gtkgui.widgets.textview import TextView
 from pynicotine.gtkgui.widgets.theme import add_css_class
 
@@ -150,7 +151,7 @@ class PluginSettings(Dialog):
 
         label = self._generate_widget_container(description, homogeneous=True)
         self.option_widgets[option_name] = combobox = ComboBox(
-            container=label.get_parent(), label=label, items=((item, item) for item in items)
+            container=label.get_parent(), label=label, items=items
         )
         self.application.preferences.set_widget(combobox, option_value)
 
@@ -195,14 +196,24 @@ class PluginSettings(Dialog):
             self.application.window, parent=scrolled_window, activate_row_callback=self.on_row_activated,
             delete_accelerator_callback=self.on_delete_accelerator,
             columns={
+                # Visible columns
                 "description": {
                     "column_type": "text",
                     "title": description
+                },
+
+                # Hidden data columns
+                "id_data": {
+                    "data_type": GObject.TYPE_INT,
+                    "default_sort_type": "ascending",
+                    "iterator_key": True
                 }
             }
         )
+        rows = [[item, index] for index, item in enumerate(option_value)]
         treeview.description = description
-        self.application.preferences.set_widget(treeview, option_value)
+        treeview.row_id = len(rows)
+        self.application.preferences.set_widget(treeview, rows)
 
         button_container = Gtk.Box(margin_end=6, margin_bottom=6, margin_start=6, margin_top=6,
                                    spacing=6, visible=True)
@@ -335,7 +346,10 @@ class PluginSettings(Dialog):
 
         from pynicotine.gtkgui.widgets.treeview import TreeView
         if isinstance(widget, TreeView):
-            return list(widget.iterators)
+            return [
+                widget.get_row_value(iterator, "description")
+                for row_id, iterator in sorted(widget.iterators.items())
+            ]
 
         if isinstance(widget, FileChooserButton):
             return widget.get_path()
@@ -358,7 +372,8 @@ class PluginSettings(Dialog):
         if not value:
             return
 
-        treeview.add_row([value])
+        treeview.row_id += 1
+        treeview.add_row([value, treeview.row_id])
 
     def on_add(self, _button, treeview):
 
@@ -378,15 +393,17 @@ class PluginSettings(Dialog):
         if not value:
             return
 
-        treeview, iterator = data
+        treeview, row_id = data
+        iterator = treeview.iterators[row_id]
 
         treeview.remove_row(iterator)
-        treeview.add_row([value])
+        treeview.add_row([value, row_id])
 
     def on_edit(self, _button=None, treeview=None):
 
         for iterator in treeview.get_selected_rows():
-            value = treeview.get_row_value(iterator, "description")
+            value = treeview.get_row_value(iterator, "description") or ""
+            row_id = treeview.get_row_value(iterator, "id_data")
 
             EntryDialog(
                 parent=self,
@@ -394,14 +411,17 @@ class PluginSettings(Dialog):
                 message=treeview.description,
                 action_button_label=_("_Edit"),
                 callback=self.on_edit_response,
-                callback_data=(treeview, iterator),
+                callback_data=(treeview, row_id),
                 default=value
             ).present()
             return
 
     def on_remove(self, _button=None, treeview=None):
-        for iterator in reversed(treeview.get_selected_rows()):
-            treeview.remove_row(iterator)
+        for iterator in reversed(list(treeview.get_selected_rows())):
+            row_id = treeview.get_row_value(iterator, "id_data")
+            orig_iterator = treeview.iterators[row_id]
+
+            treeview.remove_row(orig_iterator)
 
     def on_row_activated(self, treeview, *_args):
         self.on_edit(treeview=treeview)
@@ -414,14 +434,13 @@ class PluginSettings(Dialog):
 
     def on_ok(self, *_args):
 
+        plugin = core.pluginhandler.enabled_plugins[self.plugin_id]
+
         for name in self.plugin_settings:
             value = self._get_widget_data(self.option_widgets[name])
 
             if value is not None:
-                config.sections["plugins"][self.plugin_id.lower()][name] = value
-
-        core.pluginhandler.plugin_settings(
-            self.plugin_id, core.pluginhandler.enabled_plugins[self.plugin_id])
+                plugin.settings[name] = value
 
         self.close()
 
