@@ -23,6 +23,7 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import sys
 import time
 
 from gi.repository import Gio
@@ -57,7 +58,6 @@ from pynicotine.gtkgui.widgets.theme import set_use_header_bar
 from pynicotine.gtkgui.widgets.window import Window
 from pynicotine.logfacility import log
 from pynicotine.slskmessages import UserStatus
-from pynicotine.utils import human_speed
 from pynicotine.utils import open_folder_path
 
 
@@ -71,6 +71,8 @@ class MainWindow(Window):
         self.away_timer_id = None
         self.away_cooldown_time = 0
         self.gesture_click = None
+        self.window_active_handler = None
+        self.window_visible_handler = None
 
         # Load UI
 
@@ -107,6 +109,7 @@ class MainWindow(Window):
             self.header_end_container,
             self.header_menu,
             self.header_title,
+            self.hide_window_button,
             self.horizontal_paned,
             self.interests_container,
             self.interests_end,
@@ -115,7 +118,6 @@ class MainWindow(Window):
             self.interests_toolbar,
             self.log_container,
             self.log_search_bar,
-            self.log_search_entry,
             self.log_view_container,
             self.private_content,
             self.private_end,
@@ -219,25 +221,26 @@ class MainWindow(Window):
             self.vertical_paned.child_set_property(self.log_container, "shrink", False)
 
         # Logging
-        self.log_view = TextView(self.log_view_container, auto_scroll=not config.sections["logging"]["logcollapsed"],
-                                 parse_urls=False, editable=False, vertical_margin=5, pixels_below_lines=2)
-        self.log_search_bar = TextSearchBar(self.log_view.widget, self.log_search_bar, self.log_search_entry,
-                                            controller_widget=self.log_container)
+        self.log_view = TextView(
+            self.log_view_container, auto_scroll=not config.sections["logging"]["logcollapsed"],
+            parse_urls=False, editable=False, vertical_margin=5, pixels_below_lines=2
+        )
+        self.log_search_bar = TextSearchBar(
+            self.log_view.widget, self.log_search_bar, controller_widget=self.log_container,
+            placeholder_text=_("Search log…")
+        )
 
         self.create_log_context_menu()
         events.connect("log-message", self.log_callback)
 
         # Events
         for event_name, callback in (
-            ("schedule-quit", self.schedule_quit),
             ("server-login", self.update_user_status),
             ("server-disconnect", self.update_user_status),
             ("set-connection-stats", self.set_connection_stats),
             ("shares-preparing", self.shares_preparing),
             ("shares-ready", self.shares_ready),
             ("shares-scanning", self.shares_scanning),
-            ("update-download-limits", self.update_download_limits),
-            ("update-upload-limits", self.update_upload_limits),
             ("user-status", self.user_status)
         ):
             events.connect(event_name, callback)
@@ -342,8 +345,8 @@ class MainWindow(Window):
         self.gesture_click.connect("pressed", self.on_cancel_auto_away)
 
         # Clear notifications when main window is focused
-        self.widget.connect("notify::is-active", self.on_window_active_changed)
-        self.widget.connect("notify::visible", self.on_window_visible_changed)
+        self.window_active_handler = self.widget.connect("notify::is-active", self.on_window_active_changed)
+        self.window_visible_handler = self.widget.connect("notify::visible", self.on_window_visible_changed)
 
         # System window close (X)
         if GTK_API_VERSION >= 4:
@@ -372,7 +375,7 @@ class MainWindow(Window):
         self.set_urgency_hint(False)
 
     def on_window_visible_changed(self, *_args):
-        self.application.tray_icon.update_window_visibility()
+        self.application.tray_icon.update()
 
     def update_title(self):
 
@@ -386,17 +389,17 @@ class MainWindow(Window):
             # Private Chats have a higher priority
             user = self.privatechat.highlighted_users[-1]
             notification_text = _("Private Message from %(user)s") % {"user": user}
+            self.set_urgency_hint(True)
 
         elif self.chatrooms.highlighted_rooms:
             # Allow for the possibility the username is not available
             room, user = list(self.chatrooms.highlighted_rooms.items())[-1]
             notification_text = _("Mentioned by %(user)s in Room %(room)s") % {"user": user, "room": room}
+            self.set_urgency_hint(True)
 
         elif any(is_important for is_important in self.search.unread_pages.values()):
             notification_text = _("Wishlist Results Found")
 
-        self.set_urgency_hint(bool(notification_text))
-
         if not notification_text:
             self.set_title(pynicotine.__application_name__)
             return
@@ -463,21 +466,20 @@ class MainWindow(Window):
         self.add_action(action)
 
         action = Gio.SimpleAction(name="toggle-status")
-        action.set_enabled(False)
         action.connect("activate", self.on_toggle_status)
         self.add_action(action)
 
         # View
 
-        state = GLib.Variant("b", not config.sections["logging"]["logcollapsed"])
+        state = GLib.Variant.new_boolean(not config.sections["logging"]["logcollapsed"])
         action = Gio.SimpleAction(name="show-log-pane", state=state)
         action.connect("change-state", self.on_show_log_pane)
         self.add_action(action)
 
         # Search
 
-        action = Gio.SimpleAction(
-            name="search-mode", parameter_type=GLib.VariantType("s"), state=GLib.Variant("s", "global"))
+        state = GLib.Variant.new_string("global")
+        action = Gio.SimpleAction(name="search-mode", parameter_type=state.get_type(), state=state)
         action.connect("change-state", self.search.on_search_mode)
         self.add_action(action)
 
@@ -705,14 +707,14 @@ class MainWindow(Window):
 
         for tab_id, tab_text, tab_icon_name in (
             ("search", _("Search Files"), "system-search-symbolic"),
-            ("downloads", _("Downloads"), "document-save-symbolic"),
+            ("downloads", _("Downloads"), "folder-download-symbolic"),
             ("uploads", _("Uploads"), "emblem-shared-symbolic"),
             ("userbrowse", _("Browse Shares"), "folder-symbolic"),
             ("userinfo", _("User Profiles"), "avatar-default-symbolic"),
             ("private", _("Private Chat"), "mail-unread-symbolic"),
-            ("userlist", _("Buddies"), "contact-new-symbolic"),
+            ("userlist", _("Buddies"), "system-users-symbolic"),
             ("chatrooms", _("Chat Rooms"), "user-available-symbolic"),
-            ("interests", _("Interests"), "emblem-default-symbolic")
+            ("interests", _("Interests"), "face-smile-big-symbolic")
         ):
             tab = self.tabs[tab_id]
             self.notebook.append_page(tab.page, tab_text, focus_callback=tab.on_focus)
@@ -924,12 +926,7 @@ class MainWindow(Window):
     def update_user_status(self, *_args):
 
         status = core.users.login_status
-        is_online = (status != UserStatus.OFFLINE)
         is_away = (status == UserStatus.AWAY)
-        toggle_status_action = self.lookup_action("toggle-status")
-
-        # Action status
-        toggle_status_action.set_enabled(is_online)
 
         # Away mode
         if not is_away:
@@ -938,9 +935,6 @@ class MainWindow(Window):
             self.remove_away_timer()
 
         # Status bar
-        if core.uploads.pending_shutdown:
-            return
-
         username = core.users.login_username
         icon_name = USER_STATUS_ICON_NAMES[status]
         icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member
@@ -966,10 +960,11 @@ class MainWindow(Window):
             self.user_status_label.set_text(status_text)
 
         if self.user_status_button.get_active():
-            # Disable button toggled state without activating action
-            toggle_status_action.handler_block_by_func(self.on_toggle_status)
+            toggle_status_action = self.lookup_action("toggle-status")
+
+            toggle_status_action.set_enabled(False)
             self.user_status_button.set_active(False)
-            toggle_status_action.handler_unblock_by_func(self.on_toggle_status)
+            toggle_status_action.set_enabled(True)
 
     def user_status(self, msg):
         if msg.user == core.users.login_username:
@@ -980,6 +975,17 @@ class MainWindow(Window):
     def on_search(self, *_args):
         self.search.on_search()
 
+    def on_search_entry_changed(self, entry, *_args):
+        entry.props.secondary_icon_name = "edit-clear-symbolic" if entry.get_text() else None
+
+    def on_search_entry_icon_press(self, entry, icon_pos, *_args):
+
+        if icon_pos == Gtk.EntryIconPosition.SECONDARY:
+            entry.set_text("")
+            return
+
+        self.on_search()
+
     # User Info #
 
     def on_show_user_profile(self, *_args):
@@ -1132,48 +1138,13 @@ class MainWindow(Window):
         self.status_label.set_tooltip_text(msg)
         self.status_label.set_visible(True)
 
-    def set_connection_stats(self, total_conns=0, download_bandwidth=0, upload_bandwidth=0):
+    def set_connection_stats(self, total_conns=0, **_kwargs):
 
         total_conns_text = repr(total_conns)
-        download_bandwidth = human_speed(download_bandwidth)
-        upload_bandwidth = human_speed(upload_bandwidth)
-        download_bandwidth_text = f"{download_bandwidth} ( {len(core.downloads.active_users)} )"
-        upload_bandwidth_text = f"{upload_bandwidth} ( {len(core.uploads.active_users)} )"
 
         if self.connections_label.get_text() != total_conns_text:
             self.connections_label.set_text(total_conns_text)
 
-        if self.download_status_label.get_text() != download_bandwidth_text:
-            self.download_status_label.set_text(download_bandwidth_text)
-            self.application.tray_icon.set_download_status(_("Downloads: %(speed)s") % {"speed": download_bandwidth})
-
-        if self.upload_status_label.get_text() != upload_bandwidth_text:
-            self.upload_status_label.set_text(upload_bandwidth_text)
-            self.application.tray_icon.set_upload_status(_("Uploads: %(speed)s") % {"speed": upload_bandwidth})
-
-    def update_download_limits(self):
-        self.update_bandwidth_label_underlines(transfer_type="download")
-
-    def update_upload_limits(self):
-        self.update_bandwidth_label_underlines(transfer_type="upload")
-
-    def update_bandwidth_label_underlines(self, transfer_type):
-        """Underline status bar bandwidth labels when alternative speed limits
-        are active."""
-
-        if transfer_type == "download":
-            label = self.download_status_label
-            config_key = "use_download_speed_limit"
-        else:
-            label = self.upload_status_label
-            config_key = "use_upload_speed_limit"
-
-        if config.sections["transfers"][config_key] == "alternative":
-            add_css_class(label, "underline")
-            return
-
-        remove_css_class(label, "underline")
-
     def shares_preparing(self):
 
         # Hide widget to keep tooltips for other widgets visible
@@ -1191,18 +1162,17 @@ class MainWindow(Window):
         self.scan_progress_spinner.start()
 
     def shares_ready(self, _successful):
-
         self.scan_progress_container.set_visible(False)
         self.scan_progress_spinner.stop()
 
     def on_toggle_status(self, *_args):
 
         if core.uploads.pending_shutdown:
-            core.uploads.pending_shutdown = False
-            self.update_user_status()
-            return
+            core.uploads.cancel_shutdown()
+        else:
+            self.application.lookup_action("away").activate()
 
-        self.application.lookup_action("away").activate()
+        self.user_status_button.set_active(False)
 
     # Exit #
 
@@ -1219,22 +1189,6 @@ class MainWindow(Window):
 
         return True
 
-    def schedule_quit(self, should_finish_uploads):
-
-        if not should_finish_uploads:
-            return
-
-        icon_name = "system-shutdown-symbolic"
-        icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member
-        toggle_status_action = self.lookup_action("toggle-status")
-
-        toggle_status_action.handler_block_by_func(self.on_toggle_status)
-        self.user_status_button.set_active(True)
-        toggle_status_action.handler_unblock_by_func(self.on_toggle_status)
-
-        self.user_status_icon.set_from_icon_name(icon_name, *icon_args)
-        self.user_status_label.set_text(_("Quitting..."))
-
     def hide(self):
 
         if not self.is_visible():
@@ -1248,6 +1202,18 @@ class MainWindow(Window):
         config.write_configuration()
 
         # Hide window
+        if sys.platform == "darwin":
+            # macOS-specific way to hide the application, to ensure it is restored when clicking the dock icon
+            self.hide_window_button.set_action_name("gtkinternal.hide")
+            self.hide_window_button.emit("clicked")
+            return
+
+        if sys.platform == "win32":
+            if GTK_API_VERSION >= 4:
+                self.widget.minimize()
+            else:
+                self.widget.iconify()
+
         super().hide()
 
     def destroy(self):
@@ -1261,4 +1227,7 @@ class MainWindow(Window):
         self.popup_menu_log_view.destroy()
         self.popup_menu_log_categories.destroy()
 
+        self.widget.disconnect(self.window_active_handler)
+        self.widget.disconnect(self.window_visible_handler)
+
         super().destroy()
