@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2023 Nicotine+ Contributors
+# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
 #
 # GNU GENERAL PUBLIC LICENSE
 #    Version 3, 29 June 2007
@@ -17,6 +17,7 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import random
 import shutil
 import sys
 
@@ -55,17 +56,25 @@ if USE_COLOR_SCHEME_PORTAL:
 
     def read_color_scheme():
 
+        color_scheme = None
+
         try:
             result = SETTINGS_PORTAL.call_sync(
-                "Read", GLib.Variant("(ss)", ("org.freedesktop.appearance", "color-scheme")),
-                Gio.DBusCallFlags.NONE, -1, None
+                method_name="Read",
+                parameters=GLib.Variant.new_tuple(
+                    GLib.Variant.new_string("org.freedesktop.appearance"),
+                    GLib.Variant.new_string("color-scheme")
+                ),
+                flags=Gio.DBusCallFlags.NONE,
+                timeout_msec=1000,
+                cancellable=None
             )
-
-            return result.unpack()[0]
+            color_scheme, = result.unpack()
 
         except Exception as error:
             log.add_debug("Cannot read color scheme, falling back to GTK theme preference: %s", error)
-            return None
+
+        return color_scheme
 
     def on_color_scheme_changed(_proxy, _sender_name, signal_name, parameters):
 
@@ -82,9 +91,13 @@ if USE_COLOR_SCHEME_PORTAL:
 
     try:
         SETTINGS_PORTAL = Gio.DBusProxy.new_for_bus_sync(
-            Gio.BusType.SESSION, Gio.DBusProxyFlags.NONE, None,
-            "org.freedesktop.portal.Desktop", "/org/freedesktop/portal/desktop",
-            "org.freedesktop.portal.Settings", None
+            bus_type=Gio.BusType.SESSION,
+            flags=Gio.DBusProxyFlags.NONE,
+            info=None,
+            name="org.freedesktop.portal.Desktop",
+            object_path="/org/freedesktop/portal/desktop",
+            interface_name="org.freedesktop.portal.Settings",
+            cancellable=None
         )
         SETTINGS_PORTAL.connect("g-signal", on_color_scheme_changed)
 
@@ -131,6 +144,12 @@ def set_default_font_size():
     font_size = str(int(font_size) + 1)
     GTK_SETTINGS.props.gtk_font_name = " ".join((font_name, font_size))
 
+    if GTK_API_VERSION == 3:
+        return
+
+    # Enable OS-specific font tweaks
+    GTK_SETTINGS.props.gtk_font_rendering = Gtk.FontRendering.MANUAL  # pylint: disable=no-member
+
 
 def set_visual_settings():
 
@@ -154,26 +173,26 @@ def set_global_css():
     css = bytearray()
 
     with open(encode_path(os.path.join(css_folder_path, "style.css")), "rb") as file_handle:
-        css.extend(file_handle.read())
+        css += file_handle.read()
 
     if GTK_API_VERSION >= 4:
         add_provider_func = Gtk.StyleContext.add_provider_for_display  # pylint: disable=no-member
         display = Gdk.Display.get_default()
 
         with open(encode_path(os.path.join(css_folder_path, "style_gtk4.css")), "rb") as file_handle:
-            css.extend(file_handle.read())
+            css += file_handle.read()
 
-        if sys.platform == "darwin":
+        if sys.platform == "win32":
+            with open(encode_path(os.path.join(css_folder_path, "style_gtk4_win32.css")), "rb") as file_handle:
+                css += file_handle.read()
+
+        elif sys.platform == "darwin":
             with open(encode_path(os.path.join(css_folder_path, "style_gtk4_darwin.css")), "rb") as file_handle:
-                css.extend(file_handle.read())
+                css += file_handle.read()
 
         if LIBADWAITA_API_VERSION:
             with open(encode_path(os.path.join(css_folder_path, "style_libadwaita.css")), "rb") as file_handle:
-                css.extend(file_handle.read())
-
-            if sys.platform in {"win32", "darwin"}:
-                with open(encode_path(os.path.join(css_folder_path, "style_libadwaita_csd.css")), "rb") as file_handle:
-                    css.extend(file_handle.read())
+                css += file_handle.read()
 
         load_css(global_css_provider, css)
 
@@ -182,7 +201,7 @@ def set_global_css():
         display = Gdk.Screen.get_default()
 
         with open(encode_path(os.path.join(css_folder_path, "style_gtk3.css")), "rb") as file_handle:
-            css.extend(file_handle.read())
+            css += file_handle.read()
 
         load_css(global_css_provider, css)
 
@@ -207,11 +226,11 @@ else:
 CUSTOM_ICON_THEME_NAME = ".nicotine-icon-theme"
 FILE_TYPE_ICON_LABELS = {
     "application-x-executable-symbolic": _("Executable"),
-    "audio-x-generic-symbolic": _("Audio"),
-    "image-x-generic-symbolic": _("Image"),
+    "folder-music-symbolic": _("Audio"),
+    "folder-pictures-symbolic": _("Image"),
     "package-x-generic-symbolic": _("Archive"),
-    "text-x-generic-symbolic": _("Miscellaneous"),
-    "video-x-generic-symbolic": _("Video"),
+    "folder-documents-symbolic": _("Miscellaneous"),
+    "folder-videos-symbolic": _("Video"),
     "x-office-document-symbolic": _("Document"),
     "emblem-documents-symbolic": _("Text")
 }
@@ -247,8 +266,7 @@ def load_custom_icons(update=False):
             shutil.rmtree(icon_theme_path_encoded)
 
     except Exception as error:
-        log.add_debug("Failed to remove custom icon theme folder %(theme)s: %(error)s",
-                      {"theme": icon_theme_path, "error": error})
+        log.add_debug("Failed to remove custom icon theme folder %s: %s", (icon_theme_path, error))
         return
 
     user_icon_theme_path = config.sections["ui"]["icontheme"]
@@ -282,8 +300,7 @@ def load_custom_icons(update=False):
             file_handle.write(theme_file_contents)
 
     except Exception as error:
-        log.add_debug("Failed to enable custom icon theme %(theme)s: %(error)s",
-                      {"theme": user_icon_theme_path, "error": error})
+        log.add_debug("Failed to enable custom icon theme %s: %s", (user_icon_theme_path, error))
         return
 
     icon_names = (
@@ -363,13 +380,13 @@ def get_file_type_icon_name(basename):
     extension = extension.lower()
 
     if extension in FileTypes.AUDIO:
-        return "audio-x-generic-symbolic"
+        return "folder-music-symbolic"
 
     if extension in FileTypes.IMAGE:
-        return "image-x-generic-symbolic"
+        return "folder-pictures-symbolic"
 
     if extension in FileTypes.VIDEO:
-        return "video-x-generic-symbolic"
+        return "folder-videos-symbolic"
 
     if extension in FileTypes.ARCHIVE:
         return "package-x-generic-symbolic"
@@ -383,7 +400,7 @@ def get_file_type_icon_name(basename):
     if extension in FileTypes.EXECUTABLE:
         return "application-x-executable-symbolic"
 
-    return "text-x-generic-symbolic"
+    return "folder-documents-symbolic"
 
 
 def on_icon_theme_changed(*_args):
@@ -442,11 +459,11 @@ def remove_css_class(widget, css_class):
 def load_css(css_provider, data):
 
     try:
-        css_provider.load_from_string(data.decode("utf-8"))
+        css_provider.load_from_string(data.decode())
 
     except AttributeError:
         try:
-            css_provider.load_from_data(data.decode("utf-8"), length=-1)
+            css_provider.load_from_data(data.decode(), length=-1)
 
         except TypeError:
             css_provider.load_from_data(data)
@@ -468,7 +485,7 @@ def _get_custom_font_css():
         font_description = Pango.FontDescription.from_string(font)
 
         if font_description.get_set_fields() & (Pango.FontMask.FAMILY | Pango.FontMask.SIZE):
-            css.extend(
+            css += (
                 f"""
                 {css_selector} {{
                     font-family: '{font_description.get_family()}';
@@ -476,7 +493,7 @@ def _get_custom_font_css():
                     font-style: {PANGO_STYLES.get(font_description.get_style(), "normal")};
                     font-weight: {PANGO_WEIGHTS.get(font_description.get_weight(), "normal")};
                 }}
-                """.encode("utf-8")
+                """.encode()
             )
 
     return css
@@ -496,12 +513,12 @@ def _get_custom_color_css():
     offline_color = config.sections["ui"]["useroffline"]
 
     if _is_color_valid(online_color) and _is_color_valid(away_color) and _is_color_valid(offline_color):
-        css.extend(
+        css += (
             f"""
             .user-status {{
                 -gtk-icon-palette: success {online_color}, warning {away_color}, error {offline_color};
             }}
-            """.encode("utf-8")
+            """.encode()
         )
 
     # Text colors
@@ -512,16 +529,15 @@ def _get_custom_color_css():
         (".notebook-tab-changed", config.sections["ui"]["tab_changed"]),
         (".notebook-tab-highlight", config.sections["ui"]["tab_hilite"]),
         ("entry", config.sections["ui"]["inputcolor"]),
-        ("treeview", treeview_text_color),
-        (".search-view treeview:disabled", config.sections["ui"]["searchq"])
+        ("treeview .cell:not(:disabled):not(:selected):not(.progressbar)", treeview_text_color)
     ):
         if _is_color_valid(color):
-            css.extend(
+            css += (
                 f"""
                 {css_selector} {{
                     color: {color};
                 }}
-                """.encode("utf-8")
+                """.encode()
             )
 
     # Background colors
@@ -529,17 +545,17 @@ def _get_custom_color_css():
         ("entry", config.sections["ui"]["textbg"]),
     ):
         if _is_color_valid(color):
-            css.extend(
+            css += (
                 f"""
                 {css_selector} {{
                     background: {color};
                 }}
-                """.encode("utf-8")
+                """.encode()
             )
 
     # Reset treeview column header colors
     if treeview_text_color:
-        css.extend(
+        css += (
             b"""
             treeview header {
                 color: initial;
@@ -547,6 +563,22 @@ def _get_custom_color_css():
             """
         )
 
+    # Workaround for GTK bug where tree view colors don't update until moving the
+    # cursor over the widget. Changing the color of the text caret to a random one
+    # forces the tree view to re-render with new icon/text colors (text carets are
+    # never visible in our tree views, so usability is unaffected).
+    css += (
+        f"""
+        treeview {{
+            caret-color: #{random.randint(0, 0xFFFFFF):06x};
+        }}
+
+        treeview popover {{
+            caret-color: initial;
+        }}
+        """.encode()
+    )
+
     return css
 
 
@@ -558,10 +590,10 @@ def update_custom_css():
         .colored-icon {{
             -gtk-icon-style: {"regular" if using_custom_icon_theme else "symbolic"};
         }}
-        """.encode("utf-8")
+        """.encode()
     )
-    css.extend(_get_custom_font_css())
-    css.extend(_get_custom_color_css())
+    css += _get_custom_font_css()
+    css += _get_custom_color_css()
 
     load_css(CUSTOM_CSS_PROVIDER, css)
 
