@@ -27,29 +27,45 @@
 import os
 import time
 
-from pynicotine import slskmessages
 from pynicotine.config import config
 from pynicotine.core import core
 from pynicotine.events import events
 from pynicotine.logfacility import log
 from pynicotine.shares import PermissionLevel
+from pynicotine.slskmessages import ConnectionType
+from pynicotine.slskmessages import EmitNetworkMessageEvents
+from pynicotine.slskmessages import FileTransferInit
+from pynicotine.slskmessages import increment_token
+from pynicotine.slskmessages import initial_token
+from pynicotine.slskmessages import PlaceInQueueResponse
+from pynicotine.slskmessages import SendUploadSpeed
+from pynicotine.slskmessages import SetUploadLimit
+from pynicotine.slskmessages import TransferDirection
 from pynicotine.slskmessages import TransferRejectReason
+from pynicotine.slskmessages import TransferRequest
+from pynicotine.slskmessages import TransferResponse
+from pynicotine.slskmessages import UploadDenied
+from pynicotine.slskmessages import UploadFailed
+from pynicotine.slskmessages import UploadFile
+from pynicotine.slskmessages import UserStatus
 from pynicotine.transfers import Transfer
 from pynicotine.transfers import Transfers
 from pynicotine.transfers import TransferStatus
 from pynicotine.utils import encode_path
-from pynicotine.utils import human_speed
 
 
 class Uploads(Transfers):
+    __slots__ = ("pending_shutdown", "upload_speed", "token", "_pending_network_msgs",
+                 "_user_update_counter", "_user_update_counters", "_upload_queue_timer_id",
+                 "_retry_failed_uploads_timer_id")
 
     def __init__(self):
 
-        super().__init__(transfers_file_path=os.path.join(config.data_folder_path, "uploads.json"))
+        super().__init__(name="uploads")
 
         self.pending_shutdown = False
         self.upload_speed = 0
-        self.token = 0
+        self.token = initial_token()
 
         self._pending_network_msgs = []
         self._user_update_counter = 0
@@ -59,14 +75,16 @@ class Uploads(Transfers):
         self._retry_failed_uploads_timer_id = None
 
         for event_name, callback in (
+            ("ban-user", self._ban_user),
+            ("ban-user-ip", self._ban_user),
             ("file-connection-closed", self._file_connection_closed),
             ("file-transfer-init", self._file_transfer_init),
             ("file-upload-progress", self._file_upload_progress),
             ("peer-connection-closed", self._peer_connection_error),
             ("peer-connection-error", self._peer_connection_error),
             ("place-in-queue-request", self._place_in_queue_request),
+            ("privileged-users", self._privileged_users),
             ("queue-upload", self._queue_upload),
-            ("schedule-quit", self._schedule_quit),
             ("set-connection-stats", self._set_connection_stats),
             ("shares-ready", self._shares_ready),
             ("transfer-request", self._transfer_request),
@@ -77,20 +95,11 @@ class Uploads(Transfers):
         ):
             events.connect(event_name, callback)
 
-    def _schedule_quit(self, should_finish_uploads):
-
-        if not should_finish_uploads:
-            return
-
-        self.pending_shutdown = True
-        self._check_upload_queue()
-
     def _quit(self):
 
         super()._quit()
 
         self.upload_speed = 0
-        self.token = 0
 
     def _server_login(self, msg):
 
@@ -259,17 +268,12 @@ class Uploads(Transfers):
             if os.access(encode_path(real_path), os.R_OK):
                 return True
 
-            log.add_transfer("Cannot access file, not sharing: %(virtual_name)s with real path %(path)s", {
-                "virtual_name": virtual_path,
-                "path": real_path
-            })
+            log.add_transfer("Cannot access file, not sharing: %s with real path %s",
+                             (virtual_path, real_path))
 
         except Exception:
-            log.add_transfer(("Requested file path contains invalid characters or other errors, not sharing: "
-                              "%(virtual_name)s with real path %(path)s"), {
-                "virtual_name": virtual_path,
-                "path": real_path
-            })
+            log.add_transfer("Requested file path contains invalid characters or other errors, not sharing: "
+                             "%s with real path %s", (virtual_path, real_path))
 
         return False
 
@@ -296,45 +300,20 @@ class Uploads(Transfers):
         else:
             speed_limit = 0
 
-        core.send_message_to_network_thread(slskmessages.SetUploadLimit(speed_limit, limit_by))
+        core.send_message_to_network_thread(SetUploadLimit(speed_limit, limit_by))
         self._check_upload_queue()
 
     # Transfer Actions #
 
-    def _append_transfer(self, transfer):
+    def _enqueue_transfer(self, transfer):
 
         username = transfer.username
-        virtual_path = transfer.virtual_path
-        old_upload = self.transfers.get(username + virtual_path)
+
+        super()._enqueue_transfer(transfer)
 
         if self.is_privileged(username):
             transfer.modifier = "privileged" if username in core.users.privileged else "prioritized"
 
-        if old_upload is not None:
-            if virtual_path in self.queued_users.get(username, {}):
-                old_size = old_upload.size
-                new_size = transfer.size
-
-                if new_size != old_size:
-                    self._user_queue_sizes[username] -= old_size
-                    self._user_queue_sizes[username] += new_size
-
-                    old_upload.size = new_size
-
-                old_upload.folder_path = transfer.folder_path
-                self._update_transfer(old_upload)
-                return
-
-            if old_upload.status != TransferStatus.FINISHED:
-                transfer.current_byte_offset = old_upload.current_byte_offset
-                transfer.time_elapsed = old_upload.time_elapsed
-                transfer.time_left = old_upload.time_left
-                transfer.speed = old_upload.speed
-
-            self._clear_transfer(old_upload)
-
-        self.transfers[username + virtual_path] = transfer
-
     def _dequeue_transfer(self, transfer):
 
         username = transfer.username
@@ -344,6 +323,8 @@ class Uploads(Transfers):
         if username not in self.queued_users:
             self._user_update_counters.pop(username, None)
 
+        transfer.modifier = None
+
     def _activate_transfer(self, transfer, token):
         super()._activate_transfer(transfer, token)
         self._user_update_counters.pop(transfer.username, None)
@@ -360,20 +341,12 @@ class Uploads(Transfers):
 
         events.emit("update-upload", transfer, update_parent)
 
-    def _enqueue_limited_transfers(self, username):
-        # Not used for uploads
-        pass
-
     def _finish_transfer(self, transfer, already_exists=False):
 
         username = transfer.username
         virtual_path = transfer.virtual_path
 
-        self._deactivate_transfer(transfer)
-        self._close_file(transfer)
-
-        transfer.status = TransferStatus.FINISHED
-        transfer.current_byte_offset = transfer.size
+        super()._finish_transfer(transfer)
 
         if not self._auto_clear_transfer(transfer):
             self._update_transfer(transfer)
@@ -394,69 +367,35 @@ class Uploads(Transfers):
 
         self._check_upload_queue()
 
-    def _abort_transfer(self, transfer, denied_message=None, status=None, update_parent=True):
-
-        username = transfer.username
-        virtual_path = transfer.virtual_path
-
-        if transfer.sock is not None:
-            core.send_message_to_network_thread(slskmessages.CloseConnection(transfer.sock))
+    def _abort_transfer(self, transfer, status=None, denied_message=None, update_parent=True):
 
         if transfer.file_handle is not None:
-            self._close_file(transfer)
-
             log.add_upload(
                 _("Upload aborted, user %(user)s file %(file)s"), {
-                    "user": username,
-                    "file": virtual_path
+                    "user": transfer.username,
+                    "file": transfer.virtual_path
                 }
             )
 
-        elif denied_message and virtual_path in self.queued_users.get(username, {}):
-            core.send_message_to_peer(
-                username, slskmessages.UploadDenied(virtual_path, denied_message))
+        super()._abort_transfer(transfer, status=status, denied_message=denied_message)
+        self._update_user_counter(transfer.username)
 
-        self._deactivate_transfer(transfer)
-        self._dequeue_transfer(transfer)
-        self._unfail_transfer(transfer)
-        self._update_user_counter(username)
-
-        if not status:
-            return
-
-        transfer.status = status
-
-        if status not in {TransferStatus.FINISHED, TransferStatus.CANCELLED}:
-            self._fail_transfer(transfer)
-
-        events.emit("abort-upload", transfer, status, update_parent)
-
-    def _auto_clear_transfer(self, transfer):
-
-        if config.sections["transfers"]["autoclear_uploads"]:
-            self._clear_transfer(transfer)
-            return True
-
-        return False
+        if status:
+            events.emit("abort-upload", transfer, status, update_parent)
 
     def _clear_transfer(self, transfer, denied_message=None, update_parent=True):
 
-        self._abort_transfer(transfer, denied_message=denied_message)
+        virtual_path = transfer.virtual_path
+        username = transfer.username
 
-        log.add_transfer("Clearing upload %(path)s to user %(user)s", {
-            "path": transfer.virtual_path,
-            "user": transfer.username
-        })
+        log.add_transfer("Clearing upload %s to user %s", (virtual_path, username))
 
         try:
-            del self.transfers[transfer.username + transfer.virtual_path]
+            super()._clear_transfer(transfer, denied_message=denied_message)
 
         except KeyError:
-            log.add(("FIXME: failed to remove upload %(path)s to user %(user)s, not "
-                     "present in list"), {
-                "path": transfer.virtual_path,
-                "user": transfer.username
-            })
+            log.add("FIXME: failed to remove upload %s to user %s, not present in list",
+                    (virtual_path, username))
 
         events.emit("clear-upload", transfer, update_parent)
 
@@ -602,7 +541,7 @@ class Uploads(Transfers):
 
             username = upload_candidate.username
 
-            if slskmessages.UserStatus.OFFLINE in (core.users.login_status, core.users.statuses.get(username)):
+            if UserStatus.OFFLINE in (core.users.login_status, core.users.statuses.get(username)):
                 # Either we are offline or the user we want to upload to is
                 if self._auto_clear_transfer(upload_candidate):
                     continue
@@ -620,57 +559,34 @@ class Uploads(Transfers):
 
             if not self.is_file_readable(virtual_path, real_path):
                 self._abort_transfer(
-                    upload_candidate, denied_message=TransferRejectReason.FILE_READ_ERROR,
-                    status=TransferStatus.LOCAL_FILE_ERROR
+                    upload_candidate, status=TransferStatus.LOCAL_FILE_ERROR,
+                    denied_message=TransferRejectReason.FILE_READ_ERROR
                 )
                 continue
 
-            upload_candidate.size = self._get_current_file_size(real_path)
+            current_size = self._get_current_file_size(real_path)
+
+            if current_size is not None:
+                upload_candidate.size = current_size
+
             final_upload_candidate = upload_candidate
 
-        self.token = slskmessages.increment_token(self.token)
+        self.token = increment_token(self.token)
 
-        log.add_transfer(
-            "Checked upload queue, requesting to upload file %(file)s with token %(token)s to user %(user)s", {
-                "file": virtual_path,
-                "token": self.token,
-                "user": username
-            }
-        )
+        log.add_transfer("Checked upload queue, requesting to upload file %s with "
+                         "token %s to user %s", (virtual_path, self.token, username))
 
         self._dequeue_transfer(final_upload_candidate)
         self._unfail_transfer(final_upload_candidate)
         self._activate_transfer(final_upload_candidate, self.token)
 
         core.send_message_to_peer(
-            username, slskmessages.TransferRequest(
-                direction=slskmessages.TransferDirection.UPLOAD, token=self.token, file=virtual_path,
+            username, TransferRequest(
+                direction=TransferDirection.UPLOAD, token=self.token, file=virtual_path,
                 filesize=final_upload_candidate.size))
 
         self._update_transfer(final_upload_candidate)
 
-    def ban_users(self, users, ban_message=None):
-        """Ban a user, cancel all the user's uploads, send a 'Banned' message
-        via the transfers, and clear the transfers from the uploads list."""
-
-        if not ban_message and config.sections["transfers"]["usecustomban"]:
-            ban_message = config.sections["transfers"]["customban"]
-
-        if ban_message:
-            status = f"{TransferRejectReason.BANNED} ({ban_message})"
-        else:
-            status = TransferRejectReason.BANNED
-
-        self.clear_uploads(
-            uploads=[upload for upload in self.transfers.copy().values() if upload.username in users],
-            denied_message=status
-        )
-
-        for username in users:
-            core.network_filter.ban_user(username)
-
-        self._check_upload_queue()
-
     def enqueue_upload(self, username, virtual_path):
 
         transfer = self.transfers.get(username + virtual_path)
@@ -696,7 +612,7 @@ class Uploads(Transfers):
             self._unfail_transfer(transfer)
             transfer.size = size
 
-        if slskmessages.UserStatus.OFFLINE in (core.users.login_status, core.users.statuses.get(username)):
+        if UserStatus.OFFLINE in (core.users.login_status, core.users.statuses.get(username)):
             # Either we are offline or the user we want to upload to is
             if not self._auto_clear_transfer(transfer):
                 self._abort_transfer(transfer, status=TransferStatus.USER_LOGGED_OFF)
@@ -736,7 +652,7 @@ class Uploads(Transfers):
         for upload in uploads:
             if upload.status not in ignored_statuses:
                 self._abort_transfer(
-                    upload, denied_message=denied_message, status=status, update_parent=False)
+                    upload, status=status, denied_message=denied_message, update_parent=False)
 
         events.emit("abort-uploads", uploads, status)
 
@@ -756,6 +672,25 @@ class Uploads(Transfers):
 
         events.emit("clear-uploads", uploads, statuses)
 
+    def request_shutdown(self):
+        """Schedule a shutdown after all queued uploads have finished."""
+
+        if self.pending_shutdown:
+            return
+
+        self.pending_shutdown = True
+        self._check_upload_queue()
+
+        events.emit("uploads-shutdown-request")
+
+    def cancel_shutdown(self):
+
+        if not self.pending_shutdown:
+            return
+
+        self.pending_shutdown = False
+        events.emit("uploads-shutdown-cancel")
+
     # Events #
 
     def _shares_ready(self, _successful):
@@ -764,7 +699,7 @@ class Uploads(Transfers):
         """
 
         if self._pending_network_msgs:
-            core.send_message_to_network_thread(slskmessages.EmitNetworkMessageEvents(self._pending_network_msgs[:]))
+            core.send_message_to_network_thread(EmitNetworkMessageEvents(self._pending_network_msgs[:]))
             self._pending_network_msgs.clear()
 
     def _user_status(self, msg):
@@ -772,7 +707,14 @@ class Uploads(Transfers):
 
         username = msg.user
 
-        if msg.status == slskmessages.UserStatus.OFFLINE:
+        if username not in core.users.watched:
+            # Skip redundant status updates from users in joined rooms
+            return
+
+        if msg.status == UserStatus.OFFLINE:
+            for upload in self.failed_users.get(username, {}).copy().values():
+                self._abort_transfer(upload, status=TransferStatus.USER_LOGGED_OFF)
+
             for upload in self.active_users.get(username, {}).copy().values():
                 if upload.status == TransferStatus.TRANSFERRING:
                     continue
@@ -780,14 +722,18 @@ class Uploads(Transfers):
                 if not self._auto_clear_transfer(upload):
                     self._abort_transfer(upload, status=TransferStatus.USER_LOGGED_OFF)
 
-            for upload in self.failed_users.get(username, {}).copy().values():
-                self._abort_transfer(upload, status=TransferStatus.USER_LOGGED_OFF)
-
+            self._online_users.discard(username)
             return
 
+        # No need to check transfers on away status change
+        if username in self._online_users:
+            return
+
+        # User logged in, mark "User logged off" transfers as cancelled
         for upload in self.failed_users.get(username, {}).copy().values():
-            if upload.status == TransferStatus.USER_LOGGED_OFF:
-                self._abort_transfer(upload, status=TransferStatus.CANCELLED)
+            self._abort_transfer(upload, status=TransferStatus.CANCELLED)
+
+        self._online_users.add(username)
 
     def _user_stats(self, msg):
         """Server code 36."""
@@ -795,20 +741,66 @@ class Uploads(Transfers):
         if msg.user == core.users.login_username:
             self.upload_speed = msg.avgspeed
 
+    def _privileged_users(self, _msg):
+        """Server code 69."""
+
+        log.add_transfer("%s privileged users", len(core.users.privileged))
+
     def _set_connection_stats(self, upload_bandwidth=0, **_unused):
         self.total_bandwidth = upload_bandwidth
 
-    def _peer_connection_error(self, username, msgs=None, is_offline=False, is_timeout=True):
+    def _ban_user(self, username=None, ip_address=None):
+        """Ban a user, cancel all the user's uploads, send a 'Banned' message
+        via the transfers, and clear the transfers from the uploads list."""
 
-        if msgs is None:
+        ban_message = None
+
+        if config.sections["transfers"]["usecustomban"]:
+            ban_message = config.sections["transfers"]["customban"]
+
+        if ban_message:
+            status = f"{TransferRejectReason.BANNED} ({ban_message})"
+        else:
+            status = TransferRejectReason.BANNED
+
+        removed_uploads = []
+
+        if not username and ip_address:
+            for active_uploads in self.active_users.values():
+                for upload in active_uploads.values():
+                    active_ip_address, _port = upload.sock.getsockname()
+
+                    if active_ip_address == ip_address:
+                        username = upload.username
+                        break
+
+        for uploads in (
+            self.active_users.get(username, {}),
+            self.queued_users.get(username, {}),
+            self.failed_users.get(username, {})
+        ):
+            for upload in uploads.values():
+                removed_uploads.append(upload)
+
+        self.clear_uploads(uploads=removed_uploads, denied_message=status)
+        self._check_upload_queue()
+
+    def _peer_connection_error(self, username, conn_type, msgs, is_offline=False, is_timeout=True):
+
+        if not msgs:
             return
 
+        if conn_type not in {ConnectionType.FILE, ConnectionType.PEER}:
+            return
+
+        failed_msg_types = {TransferRequest, FileTransferInit}
+
         for msg in msgs:
-            if msg.__class__ in (slskmessages.TransferRequest, slskmessages.FileTransferInit):
+            if msg.__class__ in failed_msg_types:
                 self._cant_connect_upload(username, msg.token, is_offline, is_timeout)
 
-    def _peer_connection_closed(self, username, msgs=None):
-        self._peer_connection_error(username, msgs, is_timeout=False)
+    def _peer_connection_closed(self, username, conn_type, msgs):
+        self._peer_connection_error(username, conn_type, msgs, is_timeout=False)
 
     def _cant_connect_upload(self, username, token, is_offline, is_timeout):
         """We can't connect to the user, either way (TransferRequest,
@@ -828,13 +820,8 @@ class Uploads(Transfers):
         else:
             status = TransferStatus.CONNECTION_CLOSED
 
-        log.add_transfer(("Upload attempt for file %(filename)s with token %(token)s to user %(user)s failed "
-                          "with status %(status)s"), {
-            "filename": upload.virtual_path,
-            "token": token,
-            "user": username,
-            "status": status
-        })
+        log.add_transfer("Upload attempt for file %s with token %s to user %s failed "
+                         "with status %s", (upload.virtual_path, token, username, status))
 
         upload_cleared = is_offline and self._auto_clear_transfer(upload)
 
@@ -857,23 +844,31 @@ class Uploads(Transfers):
         real_path = core.shares.virtual2real(virtual_path)
         allowed, reason, size = self._check_queue_upload_allowed(username, msg.addr, virtual_path, real_path, msg)
 
-        log.add_transfer(("Upload request for file %(filename)s from user: %(user)s, "
-                          "allowed: %(allowed)s, reason: %(reason)s"), {
-            "filename": virtual_path,
-            "user": username,
-            "allowed": allowed,
-            "reason": reason
-        })
+        log.add_transfer("Upload request for file %s from user: %s, allowed: %s, "
+                         "reason: %s", (virtual_path, username, allowed, reason))
 
         if not allowed:
             if reason and reason != TransferRejectReason.QUEUED:
-                core.send_message_to_peer(username, slskmessages.UploadDenied(virtual_path, reason))
+                core.send_message_to_peer(username, UploadDenied(virtual_path, reason))
 
             return
 
-        transfer = Transfer(username, virtual_path, os.path.dirname(real_path), size)
+        transfer = self.transfers.get(username + virtual_path)
+        folder_path = os.path.dirname(real_path)
+
+        if transfer is not None:
+            self._unfail_transfer(transfer)
+
+            transfer.folder_path = folder_path
+            transfer.size = size
+
+            if transfer.status == TransferStatus.FINISHED:
+                transfer.current_byte_offset = transfer.last_byte_offset = None
+                transfer.speed = transfer.avg_speed = transfer.time_elapsed = transfer.time_left = 0
+        else:
+            transfer = Transfer(username, virtual_path, folder_path, size)
+            self._append_transfer(transfer)
 
-        self._append_transfer(transfer)
         self._enqueue_transfer(transfer)
         self._update_transfer(transfer)
 
@@ -887,7 +882,7 @@ class Uploads(Transfers):
 
         username = msg.username
 
-        if msg.direction != slskmessages.TransferDirection.DOWNLOAD:
+        if msg.direction != TransferDirection.DOWNLOAD:
             return
 
         response = self._transfer_request_uploads(msg)
@@ -895,11 +890,9 @@ class Uploads(Transfers):
         if response is None:
             return
 
-        log.add_transfer(("Responding to legacy upload request %(token)s for file %(filename)s "
-                          "from user %(user)s, allowed: %(allowed)s, reason: %(reason)s"), {
-            "token": response.token, "filename": msg.file, "user": username,
-            "allowed": response.allowed, "reason": response.reason
-        })
+        log.add_transfer("Responding to legacy upload request %s for file %s from user %s, "
+                         "allowed: %s, reason: %s",
+                         (response.token, msg.file, username, response.allowed, response.reason))
 
         core.send_message_to_peer(username, response)
 
@@ -915,11 +908,8 @@ class Uploads(Transfers):
         virtual_path = msg.file
         token = msg.token
 
-        log.add_transfer("Received legacy upload request %(token)s for file %(filename)s from user %(user)s", {
-            "token": token,
-            "filename": virtual_path,
-            "user": username
-        })
+        log.add_transfer("Received legacy upload request %s for file %s from user %s",
+                         (token, virtual_path, username))
 
         # Is user allowed to download?
         real_path = core.shares.virtual2real(virtual_path)
@@ -927,35 +917,49 @@ class Uploads(Transfers):
 
         if not allowed:
             if reason:
-                return slskmessages.TransferResponse(allowed=False, reason=reason, token=token)
+                return TransferResponse(allowed=False, reason=reason, token=token)
 
             return None
 
         # All checks passed, user can queue file!
-        if not self.is_new_upload_accepted() or username in self.active_users:
-            transfer = Transfer(username, virtual_path, os.path.dirname(real_path), size)
+        transfer = self.transfers.get(username + virtual_path)
+        folder_path = os.path.dirname(real_path)
 
+        if transfer is not None:
+            self._unfail_transfer(transfer)
+
+            transfer.folder_path = folder_path
+            transfer.size = size
+
+            if transfer.status == TransferStatus.FINISHED:
+                transfer.current_byte_offset = transfer.last_byte_offset = None
+                transfer.speed = transfer.avg_speed = transfer.time_elapsed = transfer.time_left = 0
+        else:
+            transfer = Transfer(username, virtual_path, folder_path, size)
             self._append_transfer(transfer)
+
+        if not self.is_new_upload_accepted() or username in self.active_users:
             self._enqueue_transfer(transfer)
             self._update_transfer(transfer)
 
             # Must be emitted after the final update to prevent inconsistent state
             core.pluginhandler.upload_queued_notification(username, virtual_path, real_path)
 
-            return slskmessages.TransferResponse(allowed=False, reason=TransferRejectReason.QUEUED, token=token)
+            return TransferResponse(allowed=False, reason=TransferRejectReason.QUEUED, token=token)
 
         # All checks passed, starting a new upload.
-        size = self._get_current_file_size(real_path)
-        transfer = Transfer(username, virtual_path, os.path.dirname(real_path), size)
+        current_size = self._get_current_file_size(real_path)
+
+        if current_size is not None:
+            transfer.size = current_size
 
-        self._append_transfer(transfer)
         self._activate_transfer(transfer, token)
         self._update_transfer(transfer)
 
         # Must be emitted after the final update to prevent inconsistent state
         core.pluginhandler.upload_queued_notification(username, virtual_path, real_path)
 
-        return slskmessages.TransferResponse(allowed=True, token=token, filesize=size)
+        return TransferResponse(allowed=True, token=token, filesize=size)
 
     def _transfer_response(self, msg):
         """Peer code 41.
@@ -967,13 +971,8 @@ class Uploads(Transfers):
         token = msg.token
         reason = msg.reason
 
-        log.add_transfer(("Received response for upload with token: %(token)s, allowed: %(allowed)s, "
-                          "reason: %(reason)s, file size: %(size)s"), {
-            "token": token,
-            "allowed": msg.allowed,
-            "reason": reason,
-            "size": msg.filesize
-        })
+        log.add_transfer("Received response for upload with token: %s, allowed: %s, "
+                         "reason: %s, file size: %s", (token, msg.allowed, reason, msg.filesize))
 
         upload = self.active_users.get(username, {}).get(token)
 
@@ -1002,7 +1001,7 @@ class Uploads(Transfers):
             self._check_upload_queue()
             return
 
-        core.send_message_to_peer(upload.username, slskmessages.FileTransferInit(token=token, is_outgoing=True))
+        core.send_message_to_peer(upload.username, FileTransferInit(token=token, is_outgoing=True))
         self._check_upload_queue()
 
     def _transfer_timeout(self, transfer):
@@ -1010,13 +1009,10 @@ class Uploads(Transfers):
         if transfer.request_timer_id is None:
             return
 
-        log.add_transfer("Upload %(filename)s with token %(token)s for user %(user)s timed out", {
-            "filename": transfer.virtual_path,
-            "token": transfer.token,
-            "user": transfer.username
-        })
+        log.add_transfer("Upload %s with token %s for user %s timed out",
+                         (transfer.virtual_path, transfer.token, transfer.username))
 
-        self._abort_transfer(transfer, status=TransferStatus.CONNECTION_TIMEOUT)
+        super()._transfer_timeout(transfer)
         self._check_upload_queue()
 
     def _upload_file_error(self, username, token, error):
@@ -1027,7 +1023,13 @@ class Uploads(Transfers):
         if upload is None:
             return
 
-        self._abort_transfer(upload, status=TransferStatus.LOCAL_FILE_ERROR)
+        if isinstance(error, ValueError):
+            status = TransferStatus.CANCELLED
+            error = f"Remote client does not support large file transfers: {error}"
+        else:
+            status = TransferStatus.LOCAL_FILE_ERROR
+
+        self._abort_transfer(upload, status=status)
 
         log.add(_("Upload I/O error: %s"), error)
         self._check_upload_queue()
@@ -1047,11 +1049,8 @@ class Uploads(Transfers):
         need_update = True
         upload_started = False
 
-        log.add_transfer("Initializing upload with token %(token)s for file %(filename)s to user %(user)s", {
-            "token": token,
-            "filename": virtual_path,
-            "user": username
-        })
+        log.add_transfer("Initializing upload with token %s for file %s to user %s",
+                         (token, virtual_path, username))
 
         real_path = core.shares.virtual2real(virtual_path)
 
@@ -1066,8 +1065,7 @@ class Uploads(Transfers):
 
         else:
             upload.file_handle = file_handle
-            upload.last_update = time.monotonic()
-            upload.start_time = upload.last_update - upload.time_elapsed
+            upload.start_time = time.monotonic() - upload.time_elapsed
 
             core.statistics.append_stat_value("started_uploads", 1)
             upload_started = True
@@ -1082,7 +1080,7 @@ class Uploads(Transfers):
 
             if upload.size > 0:
                 upload.status = TransferStatus.TRANSFERRING
-                core.send_message_to_network_thread(slskmessages.UploadFile(
+                core.send_message_to_network_thread(UploadFile(
                     sock=sock, token=token, file=file_handle, size=upload.size
                 ))
 
@@ -1097,7 +1095,7 @@ class Uploads(Transfers):
             # Must be be emitted after the final update to prevent inconsistent state
             core.pluginhandler.upload_started_notification(username, virtual_path, real_path)
 
-    def _file_upload_progress(self, username, token, offset, bytes_sent):
+    def _file_upload_progress(self, username, token, offset, bytes_sent, speed=None):
         """A file upload is in progress."""
 
         upload = self.active_users.get(username, {}).get(token)
@@ -1109,29 +1107,14 @@ class Uploads(Transfers):
             events.cancel_scheduled(upload.request_timer_id)
             upload.request_timer_id = None
 
-        current_time = time.monotonic()
-        size = upload.size
-
-        if not upload.last_byte_offset:
+        if upload.last_byte_offset is None:
             upload.last_byte_offset = offset
 
-        upload.status = TransferStatus.TRANSFERRING
-        upload.time_elapsed = current_time - upload.start_time
-        upload.current_byte_offset = current_byte_offset = (offset + bytes_sent)
-        byte_difference = current_byte_offset - upload.last_byte_offset
-
-        if byte_difference:
-            core.statistics.append_stat_value("uploaded_size", byte_difference)
-
-            if size > current_byte_offset or upload.speed <= 0:
-                upload.speed = int(max(0, byte_difference // max(0.1, current_time - upload.last_update)))
-                upload.time_left = (size - current_byte_offset) // upload.speed if upload.speed else 0
-            else:
-                upload.time_left = 0
-
-        upload.last_byte_offset = current_byte_offset
-        upload.last_update = current_time
-
+        self._update_transfer_progress(
+            upload, stat_id="uploaded_size",
+            current_byte_offset=(offset + bytes_sent) if offset is not None else None,
+            speed=speed
+        )
         self._update_transfer(upload)
 
     def _file_connection_closed(self, username, token, sock, timed_out):
@@ -1149,22 +1132,23 @@ class Uploads(Transfers):
             # We finish the upload here in case the downloading peer has a slow/limited download
             # speed and finishes later than us
 
-            if upload.speed > 0:
-                # Inform the server about the last upload speed for this transfer
-                log.add_transfer("Sending upload speed %s to the server", human_speed(upload.speed))
-                core.send_message_to_server(slskmessages.SendUploadSpeed(upload.speed))
-
             self._finish_transfer(upload)
+
+            if upload.avg_speed > 0:
+                # Inform the server about the average upload speed for this transfer
+                log.add_transfer("Sending average upload speed %s to the server", upload.avg_speed)
+                core.send_message_to_server(SendUploadSpeed(upload.avg_speed))
+
             return
 
-        if core.users.statuses.get(upload.username) == slskmessages.UserStatus.OFFLINE:
+        if core.users.statuses.get(upload.username) == UserStatus.OFFLINE:
             status = TransferStatus.USER_LOGGED_OFF
         else:
             status = TransferStatus.CANCELLED
 
             # Transfer ended abruptly. Tell the peer to re-queue the file. If the transfer was
             # intentionally cancelled, the peer should ignore this message.
-            core.send_message_to_peer(upload.username, slskmessages.UploadFailed(upload.virtual_path))
+            core.send_message_to_peer(upload.username, UploadFailed(upload.virtual_path))
 
         if not self._auto_clear_transfer(upload):
             self._abort_transfer(upload, status=status)
@@ -1212,7 +1196,7 @@ class Uploads(Transfers):
 
         if queue_position > 0:
             core.send_message_to_peer(
-                username, slskmessages.PlaceInQueueResponse(virtual_path, queue_position))
+                username, PlaceInQueueResponse(virtual_path, queue_position))
 
         # Update queue position in our list of uploads
         upload.queue_position = queue_position
